--[[ ArenaLive Spectator Functions: Spectated Unit Cache Handler
Created by: Vadrak
Creation Date: 14.02.2015
Last Update: "
Description: This is no actual ArenaLive handler, just a frame that is attached to the
			 ArenaLiveSpectator frame. It will track unit info for spectateda and
			 spectatedb unit types, because the COMMENTATOR_PLAYER_UPDATE event fires
			 before player info is available for these unit types. So this handler will 
			 track unit info via an OnUpdate script and fires a custom event 
			 AL_SPEC_PLAYER_UPDATE once all unit information is available to the client.
]]--

-- Addon Name and localisation table:
local addonName, L = ...;

--[[
**************************************************
******* GENERAL HANDLER SET UP STARTS HERE *******
**************************************************
]]--
ArenaLiveSpectator.UnitCache = CreateFrame("Frame", "ArenaLiveSpectatorUnitCache");
local UnitCache = ArenaLiveSpectator.UnitCache;
local unitsWaitingForInfo = {};
local unitCache = {};
local GUIDToInfo = {};
local ON_UPDATE_THROTTLE = 0.1;
local ELAPSED = 0;

UnitCache:RegisterEvent("UPDATE_BATTLEFIELD_STATUS");



--[[
****************************************
****** HANDLER METHODS START HERE ******
****************************************
]]--
function UnitCache:Enable(statusID)
	self.statusID = statusID;
	self:RegisterEvent("COMMENTATOR_PLAYER_UPDATE");
	self:RegisterEvent("UNIT_PET");
	self:Show();
	self.enabled = true;
	UnitCache:UpdateNumPlayers();
	ArenaLive:Message("Enabling Spectated Unit Cache", "debug");
end

function UnitCache:Disable()
	self.statusID = nil;
	self:UnregisterEvent("COMMENTATOR_PLAYER_UPDATE");
	self:RegisterEvent("UNIT_PET");
	self:Hide();
	
	-- Reset unit info:
	table.wipe(unitCache);
	table.wipe(GUIDToInfo);
	table.wipe(unitsWaitingForInfo);
	self.enabled = false;
	ArenaLive:Message("Disabling Spectated Unit Cache", "debug");
end

function UnitCache:IsEnabled()
	return self.enabled;
end

function UnitCache:GetUnitInfoByGUID(guid)
	if ( guid and GUIDToInfo[guid] ) then
		return true, GUIDToInfo[guid].unit, guid, GUIDToInfo[guid].name, GUIDToInfo[guid].team;
	else
		return nil;
	end
end

function UnitCache:GetUnitInfo(unit)
	return self:GetUnitInfoByGUID(unitCache[unit]);
end

function UnitCache:GetUnitInfoByName(name)
	local GUID;
	for guid, infoTable in pairs(GUIDToInfo) do
		if ( name == infoTable.name ) then
			GUID = guid;
			break;
		end
	end
	
	return self:GetUnitInfoByGUID(GUID);
end

function UnitCache:UpdateNumPlayers()
	local numTeamA = C_Commentator.GetNumPlayers(2);
	local numTeamB = C_Commentator.GetNumPlayers(1);
	
	-- Reset old info:
	table.wipe(unitCache);
	table.wipe(GUIDToInfo);
	table.wipe(unitsWaitingForInfo);
	
	-- Get max iterations:
	local iMax;
	if ( numTeamA > numTeamB ) then
		iMax = numTeamA;
	else
		iMax = numTeamB;
	end
	
	if ( iMax > 0 ) then
		-- Iterate through units:
		local unit, exists;
		for i = 1, iMax do
			unit = "spectateda"..i;
			if ( i <= numTeamA ) then
				unitsWaitingForInfo[unit] = true;
				unit = "spectatedpeta"..i;
				unitsWaitingForInfo[unit] = UnitExists(unit) or nil;
				--[[ NOTE:	UnitExists returns false for an non-existing
							unit, that would cause the OnUpdate function
							to search for the pet units all the time.
							"or nil" prevents that. ]]
			end
			
			unit = "spectatedb"..i;
			if ( i <= numTeamB ) then
				unitsWaitingForInfo[unit] = true;
				unit = "spectatedpeta"..i;
				unitsWaitingForInfo[unit] = UnitExists(unit) or nil;
			end
		end
		
		ELAPSED = 0;
		self:Show();
	else
		ELAPSED = 0;
		self:Hide();
	end
end

function UnitCache:SetUnitInfo(unit)
	local name, realm = UnitName(unit);
	local exists = UnitExists(unit);
	local guid = UnitGUID(unit);
	if ( name and name ~= "Unknown" and exists and guid ) then
		if ( not realm ) then
			realm = GetRealmName();
		end
		name = name.."-"..realm;
		local unitType = string.match(unit, "^([a-z]+)[0-9]+$") or unit;
		local teamID;
		local database = ArenaLive:GetDBComponent(addonName);
		if ( unitType == "spectateda" or unitType == "spectatedpeta" ) then
			teamID = 1; -- (Alliance or yellow team);
		elseif ( unitType == "spectatedb"  or unitType == "spectatedpetb" ) then
			teamID = 0; -- (Horde or green team);
		end
		
		unitCache[unit] = guid;
		GUIDToInfo[guid] = {};
		GUIDToInfo[guid].unit = unit;
		GUIDToInfo[guid].guid = guid;
		GUIDToInfo[guid].name = name;
		GUIDToInfo[guid].team = teamID;
		--ArenaLive:Message("Called UnitCache:SetUnitInfo() for unit %s: exists = %s, guid = %s, teamID = %d", "debug", unit, tostring(exists), tostring(guid), teamID);
	elseif ( not exists and unitCache[unit] ) then
		-- This is only the case if we update a pet unit:
		local guid = unitCache[unit];
		local name = GUIDToInfo[guid].name;
		unitCache[unit] = nil;
		GUIDToInfo[guid] = nil;
		unitsWaitingForInfo[unit] = nil;
	end
end

local function debugNumTeamMembers()
	local numTeam1 = C_Commentator.GetNumPlayers(1);
	local numTeam2 = C_Commentator.GetNumPlayers(2);
	ArenaLive:Message("numTeam1 = %d; numTeam2 = %d", "normal", numTeam1, numTeam2);
end

function UnitCache:OnEvent(event, ...)
	if ( event == "COMMENTATOR_PLAYER_UPDATE" ) then
		debugNumTeamMembers();
		self:UpdateNumPlayers();
	elseif ( event == "UNIT_PET" ) then
		local unit = ...;
		local unitType = string.match(unit, "^([a-z]+)[0-9]+$") or unit;
		if ( unitType == "spectateda" or unitType == "spectatedb" ) then
			local number = string.match(unit, "^[a-z]+([0-9]+)$");
			local unit;
			if ( unitType == "spectateda" ) then
				unit = "spectatedpeta"..number;
			else
				unit = "spectatedpetb"..number;
			end
			
			UnitCache:SetUnitInfo(unit);
			ArenaLive:TriggerEvent("AL_SPEC_PLAYER_UPDATE", unit);
		end
	elseif ( event == "UPDATE_BATTLEFIELD_STATUS" ) then
		local statusID = ...;
		local status = GetBattlefieldStatus(statusID);
		local isSpectator = IsSpectator();
		if( status == "active" and isSpectator and ( not self.enabled or statusID ~= self.statusID ) ) then
			debugNumTeamMembers();
			self:Enable(statusID);
		elseif ( status == "none" and statusID == self.statusID ) then
			self:Disable();
		end
	end
end

function UnitCache:OnUpdate(elapsed)
	if ( self.enabled ) then
		ELAPSED = ELAPSED + elapsed;
		if ( ELAPSED >= ON_UPDATE_THROTTLE ) then
			ELAPSED = 0;
			local infoChanged;
			for unit, name in pairs(unitsWaitingForInfo) do
				local unitName = UnitName(unit);
				local exists = UnitExists(unit);
				if ( type(name) == "boolean" and unitName and unitName ~= "Unknown" and exists ) then
					UnitCache:SetUnitInfo(unit);
					if ( unitCache[unit] ) then
						unitsWaitingForInfo[unit] = nil;
					end
				end
			end
			
			if ( not next(unitsWaitingForInfo) ) then
				-- All changes were made, trigger custom event and hide frame to safe performance:
				ArenaLive:Message("All player info gathered, firing AL_SPEC_PLAYER_UPDATE event and hiding frame.", "debug");
				self:Hide();
				ArenaLive:TriggerEvent("AL_SPEC_PLAYER_UPDATE");
			end
		end
	end
end

-- Disable on initial load:
UnitCache:Disable();
UnitCache:SetScript("OnEvent", UnitCache.OnEvent);
UnitCache:SetScript("OnUpdate", UnitCache.OnUpdate);